/*
 * EpsilonNFA.h
 *
 * This file is part of Algorithms library toolkit.
 * Copyright (C) 2017 Jan Travnicek (jan.travnicek@fit.cvut.cz)

 * Algorithms library toolkit is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * Algorithms library toolkit is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Algorithms library toolkit.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef EPSILON_NFA_H_
#define EPSILON_NFA_H_

#include <ostream>
#include <sstream>

#include <alib/multimap>
#include <alib/set>

#include <core/components.hpp>
#include <common/createUnique.hpp>

#include <common/DefaultStateType.h>
#include <common/DefaultSymbolType.h>
#include <common/symbol_or_epsilon.hpp>

#include <label/InitialStateLabel.h>

#include <automaton/AutomatonException.h>

#include <core/normalize.hpp>
#include <alphabet/common/SymbolNormalize.h>
#include <automaton/common/AutomatonNormalize.h>

#include "MultiInitialStateNFA.h"
#include "NFA.h"
#include "DFA.h"

namespace automaton {

class InputAlphabet;
class States;
class FinalStates;
class InitialState;

/**
 * \brief
 * Epsilon nondeterministic finite automaton. Accepts regular languages.

 * \details
 * Definition is classical definition of finite automata.
 * A = (Q, T, I, \delta, F),
 * Q (States) = nonempty finite set of states,
 * T (TerminalAlphabet) = finite set of terminal symbols - having this empty won't let automaton do much though,
 * I (InitialState) = initial state,
 * \delta = transition function of the form A \times a -> P(Q), where A \in Q, a \in T \cup \{\epsilon\}, and P(Q) is a powerset of states,
 * F (FinalStates) = set of final states
 *
 * \tparam SymbolTypeT used for the terminal alphabet
 * \tparam StateTypeT used to the states, and the initial state of the automaton.
 */
template < class SymbolTypeT = DefaultSymbolType, class StateTypeT = DefaultStateType >
class EpsilonNFA final : public ext::CompareOperators < EpsilonNFA < SymbolTypeT, StateTypeT > >, public core::Components < EpsilonNFA < SymbolTypeT, StateTypeT >, ext::set < SymbolTypeT >, component::Set, InputAlphabet, ext::set < StateTypeT >, component::Set, std::tuple < States, FinalStates >, StateTypeT, component::Value, InitialState > {
public:
	typedef SymbolTypeT SymbolType;
	typedef StateTypeT StateType;

private:
	/**
	 * Transition function as mapping from a state times an input symbol or epsilon on the left hand side to a set of states.
	 */
	ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > transitions;

public:
	/**
	 * \brief Creates a new instance of the automaton with a concrete initial state.
	 *
	 * \param initialState the initial state of the automaton
	 */
	explicit EpsilonNFA ( StateType initialState );

	/**
	 * \brief Creates a new instance of the automaton with a concrete set of states, input alphabet, initial state, and a set of final states.
	 *
	 * \param states the initial set of states of the automaton
	 * \param inputAlphabet the initial input alphabet
	 * \param initialState the initial state of the automaton
	 * \param finalStates the initial set of final states of the automaton
	 */
	explicit EpsilonNFA ( ext::set < StateType > states, ext::set < SymbolType > inputAlphabet, StateType initialState, ext::set < StateType > finalStates );

	/*
	 * \brief Creates a new instance of Nondeterministic finite automaton based on the Deterministic finite automaton.
	 *
	 * \param other the Nondeterministic finite automaton with multiple initial states
	 */
	explicit EpsilonNFA ( const MultiInitialStateNFA < SymbolType, StateType > & other );

	/*
	 * \brief Creates a new instance of Nondeterministic finite automaton based on the Deterministic finite automaton.
	 *
	 * \param other the Nondeterministic finite automaton
	 */
	explicit EpsilonNFA ( const NFA < SymbolType, StateType > & other );

	/*
	 * \brief creates a new instance of the automaton based on the deterministic finite automaton.
	 *
	 * \param other the deterministic finite automaton
	 */
	explicit EpsilonNFA ( const DFA < SymbolType, StateType > & other );

	/**
	 * Getter of the initial state.
	 *
	 * \returns the initial state of the automaton
	 */
	const StateType & getInitialState ( ) const & {
		return this->template accessComponent < InitialState > ( ).get ( );
	}

	/**
	 * Getter of the initial state.
	 *
	 * \returns the initial state of the automaton
	 */
	StateType && getInitialState ( ) && {
		return std::move ( this->template accessComponent < InitialState > ( ).get ( ) );
	}

	/**
	 * Setter of the initial state.
	 *
	 * \param state new initial state of the automaton
	 *
	 * \returns true if the initial state was indeed changed
	 */
	bool setInitialState ( StateType state ) {
		return this->template accessComponent < InitialState > ( ).set ( std::move ( state ) );
	}

	/**
	 * Getter of states.
	 *
	 * \returns the states of the automaton
	 */
	const ext::set < StateType > & getStates ( ) const & {
		return this->template accessComponent < States > ( ).get ( );
	}

	/**
	 * Getter of states.
	 *
	 * \returns the states of the automaton
	 */
	ext::set < StateType > && getStates ( ) && {
		return std::move ( this->template accessComponent < States > ( ).get ( ) );
	}

	/**
	 * Adder of a state.
	 *
	 * \param state the new state to be added to a set of states
	 *
	 * \returns true if the state was indeed added
	 */
	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	/**
	 * Setter of states.
	 *
	 * \param states completely new set of states
	 */
	void setStates ( ext::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	/**
	 * Remover of a state.
	 *
	 * \param state a state to be removed from a set of states
	 *
	 * \returns true if the state was indeed removed
	 */
	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	/**
	 * Getter of final states.
	 *
	 * \returns the final states of the automaton
	 */
	const ext::set < StateType > & getFinalStates ( ) const & {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	/**
	 * Getter of final states.
	 *
	 * \returns the final states of the automaton
	 */
	ext::set < StateType > && getFinalStates ( ) && {
		return std::move ( this->template accessComponent < FinalStates > ( ).get ( ) );
	}

	/**
	 * Adder of a final state.
	 *
	 * \param state the new state to be added to a set of final states
	 *
	 * \returns true if the state was indeed added
	 */
	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	/**
	 * Setter of final states.
	 *
	 * \param states completely new set of final states
	 */
	void setFinalStates ( ext::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	/**
	 * Remover of a final state.
	 *
	 * \param state a state to be removed from a set of final states
	 *
	 * \returns true if the state was indeed removed
	 */
	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	/**
	 * Getter of the input alphabet.
	 *
	 * \returns the input alphabet of the automaton
	 */
	const ext::set < SymbolType > & getInputAlphabet ( ) const & {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	/**
	 * Getter of the input alphabet.
	 *
	 * \returns the input alphabet of the automaton
	 */
	ext::set < SymbolType > && getInputAlphabet ( ) && {
		return std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) );
	}

	/**
	 * Adder of a input symbol.
	 *
	 * \param symbol the new symbol to be added to an input alphabet
	 *
	 * \returns true if the symbol was indeed added
	 */
	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	/**
	 * Adder of input symbols.
	 *
	 * \param symbols new symbols to be added to an input alphabet
	 */
	void addInputSymbols ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	/**
	 * Setter of input alphabet.
	 *
	 * \param symbols completely new input alphabet
	 */
	void setInputAlphabet ( ext::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	/**
	 * Remover of an input symbol.
	 *
	 * \param symbol a symbol to be removed from an input alphabet
	 *
	 * \returns true if the symbol was indeed removed
	 */
	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * \brief Add a transition to the automaton.
	 *
	 * \details The transition is in a form A \times a -> B, where A, B \in Q and a \in T \cup \{\epsilon\}
	 *
	 * \param current the source state (A)
	 * \param input the input symbol or epsilon (a)
	 * \param next the target state (B)
	 *
	 * \throws AutomatonException when transition contains state or symbol not present in the automaton components
	 *
	 * \returns true if the transition was indeed added
	 */
	bool addTransition ( StateType from, common::symbol_or_epsilon < SymbolType > input, StateType to );

	/**
	 * \brief Add a transition to the automaton.
	 *
	 * \details The transition is in a form A \times a -> B, where A, B \in Q and a \in T
	 *
	 * \param current the source state (A)
	 * \param input the input symbol (a)
	 * \param next the target state (B)
	 *
	 * \throws AutomatonException when transition contains state or symbol not present in the automaton components
	 *
	 * \returns true if the transition was indeed added
	 */
	bool addTransition ( StateType from, SymbolType input, StateType to );

	/**
	 * \brief Add a transition to the automaton.
	 *
	 * \details The transition is in a form A \times \epsilon -> B, where A, B \in Q
	 *
	 * \param current the source state (A)
	 * \param next the target state (B)
	 *
	 * \throws AutomatonException when transition contains state or symbol not present in the automaton components
	 *
	 * \returns true if the transition was indeed added
	 */
	bool addTransition ( StateType from, StateType to );

	/**
	 * \brief Removes a transition to the automaton.
	 *
	 * \details The transition is in a form A \times a -> B, where A, B \in Q and a \in T \cup \{\epsilon\}
	 *
	 * \param current the source state (A)
	 * \param input the input symbol or epsilon (a)
	 * \param next the target state (B)
	 *
	 * \throws AutomatonException when transition contains state or symbol not present in the automaton components
	 *
	 * \returns true if the transition was indeed added
	 */
	bool removeTransition ( StateType from, common::symbol_or_epsilon < SymbolType > input, StateType to );

	/**
	 * \brief Removes a transition from the automaton.
	 *
	 * \details The transition is in a form A \times a -> B, where A, B \in Q and a \in T
	 *
	 * \param current the source state (A)
	 * \param input the input symbol (a)
	 * \param next the target state (B)
	 *
	 * \returns true if the transition was indeed removed
	 */
	bool removeTransition ( const StateType & from, const SymbolType & input, const StateType & to );

	/**
	 * \brief Removes a transition from the automaton.
	 *
	 * \details The transition is in a form A \times \epsilon -> B, where A, B \in Q
	 *
	 * \param current the source state (A)
	 * \param next the target state (B)
	 *
	 * \returns true if the transition was indeed removed
	 */
	bool removeTransition ( const StateType & from, const StateType & to );

	/**
	 * Get the transition function of the automaton in its natural form.
	 *
	 * \returns transition function of the automaton
	 */
	const ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & getTransitions ( ) const &;

	/**
	 * Get the transition function of the automaton in its natural form.
	 *
	 * \returns transition function of the automaton
	 */
	ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > && getTransitions ( ) &&;

	/**
	 * Get the epsilon transitions of the automaton
	 *
	 * \returns epsilon transitions of the automaton
	 */
	ext::multimap < StateType, StateType > getEpsilonTransitions ( ) const;

	/**
	 * Get the non-epsilon transitions of the automaton
	 *
	 * \returns non-epsilon transitions of the automaton
	 */
	ext::multimap < ext::pair < StateType, SymbolType >, StateType > getSymbolTransitions ( ) const;

	/**
	 * Get a subset of the transition function of the automaton, with the source state fixed in the transitions natural representation.
	 *
	 * \param from filter the transition function based on this state as a source state
	 *
	 * \returns a subset of the transition function of the automaton with the source state fixed
	 */
	ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > getTransitionsFromState ( const StateType & from ) const;

	/**
	 * Get a subset of epsilon transitions of the automaton, with the source state fixed in the transitions natural representation.
	 *
	 * \param from filter the transition function based on this state as a source state
	 *
	 * \returns a subset of epsilon transitions of the automaton with the source state fixed
	 */
	ext::multimap < StateType, StateType > getEpsilonTransitionsFromState ( const StateType & from ) const;

	/**
	 * Get a subset of non-epsilon transitions of the automaton, with the source state fixed in the transitions natural representation.
	 *
	 * \param from filter the transition function based on this state as a source state
	 *
	 * \returns a subset of non-epsilon transitions of the automaton with the source state fixed
	 */
	ext::multimap < ext::pair < StateType, SymbolType >, StateType > getSymbolTransitionsFromState ( const StateType & from ) const;

	/**
	 * Get a subset of the transition function of the automaton, with the target state fixed in the transitions natural representation.
	 *
	 * \param to filter the transition function based on this state as a source state
	 *
	 * \returns a subset of the transition function of the automaton with the target state fixed
	 */
	ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > getTransitionsToState ( const StateType & to ) const;

	/**
	 * Get a subset of epsilon transitions of the automaton, with the target state fixed in the transitions natural representation.
	 *
	 * \param to filter the transition function based on this state as a source state
	 *
	 * \returns a subset of epsilon transitions of the automaton with the target state fixed
	 */
	ext::multimap < StateType, StateType > getEpsilonTransitionsToState ( const StateType & to ) const;

	/**
	 * Get a subset of non-epsilon transitions of the automaton, with the target state fixed in the transitions natural representation.
	 *
	 * \param to filter the transition function based on this state as a source state
	 *
	 * \returns a subset of non-epsilon transitions of the automaton with the target state fixed
	 */
	ext::multimap < ext::pair < StateType, SymbolType >, StateType > getSymbolTransitionsToState ( const StateType & to ) const;

	/**
	 * \brief Determines whether the automaton is without epsilon transitions.
	 *
	 * \return true when automaton doesn't contain epsilon transitions, false otherwise
	 */
	bool isEpsilonFree ( ) const;

	/**
	 * \brief Determines whether the automaton is deterministic
	 *
	 * The automaton is deterministic if and only if:
	 * \li \c is epsilon free.
	 * \li \c size of transition function \delta (from state, input symbol) \leq 1
	 *
	 * \return true if the automaton is deterministic, false otherwise
	 */
	bool isDeterministic ( ) const;

	/**
	 * \brief Determines whether the automaton is total deterministic
	 *
	 * The automaton is deterministic if and only if:
	 * \li \c it is deterministic
	 * \li \c size of transition function \forall states and \forall input symbols \delta (from state, input symbol) = 1
	 *
	 * \return true if the automaton is total deterministic, false otherwise
	 */
	bool isTotal ( ) const;

	/**
	 * The actual compare method
	 *
	 * \param other the other instance
	 *
	 * \returns the actual relation between two by type same automata instances
	 */
	int compare ( const EpsilonNFA & other ) const;

	/**
	 * Print this object as raw representation to ostream.
	 *
	 * \param out ostream where to print
	 * \param instance object to print
	 *
	 * \returns modified output stream
	 */
	friend std::ostream & operator << ( std::ostream & out, const EpsilonNFA & instance ) {
		return out << "(EpsilonNFA "
			   << "states = " << instance.getStates ( )
			   << "inputAlphabet = " << instance.getInputAlphabet ( )
			   << "initialState = " << instance.getInitialState ( )
			   << "finalStates = " << instance.getFinalStates ( )
			   << "transitions = " << instance.getTransitions ( )
			   << ")";
	}

	/**
	 * Casts this instance to as compact as possible string representation.
	 *
	 * \returns string representation of the object
	 */
	explicit operator std::string ( ) const;
};

/**
 * Trait to detect whether the type parameter T is or is not EpsilonNFA. Derived from std::false_type.
 *
 * \tparam T the tested type parameter
 */
template < class T >
class isEpsilonNFA_impl : public std::false_type {};

/**
 * Trait to detect whether the type parameter T is or is not EpsilonNFA. Derived from std::true_type.
 *
 * Specialisation for EpsilonNFA.
 *
 * \tparam SymbolType used for the terminal alphabet of the automaton
 * \tparam StateType used for the terminal alphabet of the automaton
 */
template < class SymbolType, class StateType >
class isEpsilonNFA_impl < EpsilonNFA < SymbolType, StateType > > : public std::true_type {};

/**
 * Constexpr true if the type parameter T is EpsilonNFA, false otherwise.
 *
 * \tparam T the tested type parameter
 */
template < class T >
constexpr bool isEpsilonNFA = isEpsilonNFA_impl < T > { };

template<class SymbolType, class StateType >
EpsilonNFA < SymbolType, StateType >::EpsilonNFA ( ext::set < StateType > states, ext::set < SymbolType > inputAlphabet, StateType initialState, ext::set < StateType > finalStates ) : core::Components < EpsilonNFA, ext::set < SymbolType >, component::Set, InputAlphabet, ext::set < StateType >, component::Set, std::tuple < States, FinalStates >, StateType, component::Value, InitialState > ( std::move ( inputAlphabet ), std::move ( states ), std::move ( finalStates ), std::move ( initialState ) ) {
}

template<class SymbolType, class StateType >
EpsilonNFA < SymbolType, StateType >::EpsilonNFA ( StateType initialState ) : EpsilonNFA ( ext::set < StateType > { initialState }, ext::set < SymbolType > { }, initialState, ext::set < StateType > { } ) {
}


template<class SymbolType, class StateType >
EpsilonNFA < SymbolType, StateType >::EpsilonNFA ( const MultiInitialStateNFA < SymbolType, StateType > & other ) : EpsilonNFA ( other.getStates ( ) + ext::set < StateType > { common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), other.getStates ( ) ) }, other.getInputAlphabet ( ), common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), other.getStates ( ) ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) )
		transitions.insert ( ext::make_pair ( transition.first.first, common::symbol_or_epsilon < SymbolType > ( transition.first.second ) ), transition.second );

	for ( const auto & initialState : other.getInitialStates ( ) )
		transitions.insert ( ext::make_pair ( getInitialState ( ), common::symbol_or_epsilon < SymbolType > ( ) ), initialState );
}


template<class SymbolType, class StateType >
EpsilonNFA < SymbolType, StateType >::EpsilonNFA ( const NFA < SymbolType, StateType > & other ) : EpsilonNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) )
		transitions.insert ( ext::make_pair ( transition.first.first, common::symbol_or_epsilon < SymbolType > ( transition.first.second ) ), transition.second );
}

template<class SymbolType, class StateType >
EpsilonNFA < SymbolType, StateType >::EpsilonNFA ( const DFA < SymbolType, StateType > & other ) : EpsilonNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) )
		transitions.insert ( ext::make_pair ( transition.first.first, common::symbol_or_epsilon < SymbolType > ( transition.first.second ) ), transition.second );
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::addTransition ( StateType from, common::symbol_or_epsilon < SymbolType > input, StateType to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist." );

	if ( ! input.is_epsilon ( ) && !getInputAlphabet ( ).count ( input.getSymbol ( ) ) )
		throw AutomatonException ( "Input symbol \"" + ext::to_string ( input ) + "\" doesn't exist." );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist." );

	auto upper_bound = transitions.upper_bound ( ext::tie ( from, input ) );
	auto lower_bound = transitions.lower_bound ( ext::tie ( from, input ) );
	auto iter = std::lower_bound ( lower_bound, upper_bound, to, [ ] ( const auto & transition, const auto & target ) { return transition.second < target; } );
	if ( iter != upper_bound && to >= iter->second )
		return false;

	ext::pair < StateType, common::symbol_or_epsilon < SymbolType > > key = ext::make_pair ( std::move ( from ), std::move ( input ) );
	transitions.insert ( iter, std::make_pair ( std::move ( key ), std::move ( to ) ) );
	return true;
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::addTransition ( StateType from, SymbolType input, StateType to ) {
	common::symbol_or_epsilon < SymbolType > inputVariant ( std::move ( input ) );

	return addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::addTransition ( StateType from, StateType to ) {
	auto inputVariant = common::symbol_or_epsilon < SymbolType > ( );

	return addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::removeTransition ( StateType from, common::symbol_or_epsilon < SymbolType > input, StateType to ) {
	auto upper_bound = transitions.upper_bound ( ext::tie ( from, input ) );
	auto lower_bound = transitions.lower_bound ( ext::tie ( from, input ) );
	auto iter = std::find_if ( lower_bound, upper_bound, [ & ] ( const auto & transition ) { return transition.second == to; } );
	if ( iter == upper_bound )
		return false;

	transitions.erase ( iter );
	return true;
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::removeTransition ( const StateType & from, const SymbolType & input, const StateType & to ) {
	common::symbol_or_epsilon < SymbolType > inputVariant ( std::move ( input ) );

	return addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::removeTransition ( const StateType & from, const StateType & to ) {
	auto inputVariant = common::symbol_or_epsilon < SymbolType > ( );

	return addTransition ( std::move ( from ), std::move ( inputVariant ), std::move ( to ) );
}

template<class SymbolType, class StateType >
const ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & EpsilonNFA < SymbolType, StateType >::getTransitions ( ) const & {
	return transitions;
}

template<class SymbolType, class StateType >
ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > && EpsilonNFA < SymbolType, StateType >::getTransitions ( ) && {
	return std::move ( transitions );
}

template<class SymbolType, class StateType >
ext::multimap < StateType, StateType > EpsilonNFA < SymbolType, StateType >::getEpsilonTransitions ( ) const {
	ext::multimap < StateType, StateType > result;

	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions )
		if ( transition.first.second.is_epsilon ( ) )
			result.insert ( transition.first.first, transition.second );

	return result;
}

template<class SymbolType, class StateType >
ext::multimap < ext::pair < StateType, SymbolType >, StateType > EpsilonNFA < SymbolType, StateType >::getSymbolTransitions ( ) const {
	ext::multimap < ext::pair < StateType, SymbolType >, StateType > result;

	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions )
		if ( ! transition.first.second.is_epsilon ( ) )
			result.insert ( ext::make_pair ( transition.first.first, transition.first.second.getSymbol ( ) ), transition.second );

	return result;
}

template<class SymbolType, class StateType >
ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > EpsilonNFA < SymbolType, StateType >::getTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist" );

	ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > transitionsFromState;

	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState.insert ( transition );

	return transitionsFromState;
}

template<class SymbolType, class StateType >
ext::multimap < StateType, StateType > EpsilonNFA < SymbolType, StateType >::getEpsilonTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist" );

	ext::pair < StateType, common::symbol_or_epsilon < SymbolType > > key ( from, common::symbol_or_epsilon < SymbolType > ( ) );
	ext::multimap < StateType, StateType > res;
	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions.equal_range ( key ) )
		res.insert ( transition.first.first, transition.second );

	return res;
}

template<class SymbolType, class StateType >
ext::multimap < ext::pair < StateType, SymbolType >, StateType > EpsilonNFA < SymbolType, StateType >::getSymbolTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + ext::to_string ( from ) + "\" doesn't exist" );

	ext::multimap < ext::pair < StateType, SymbolType >, StateType > transitionsFromState;

	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions )
		if ( ( transition.first.first == from ) && ! transition.first.second.is_epsilon ( ) )
			transitionsFromState.insert ( std::make_pair ( ext::make_pair ( transition.first.first, transition.first.second.getSymbol ( ) ), transition.second ) );

	return transitionsFromState;
}

template<class SymbolType, class StateType >
ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > EpsilonNFA < SymbolType, StateType >::getTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist" );

	ext::multimap < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > transitionsToState;

	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions )
		if ( transition.second == to )
			transitionsToState.insert ( transition );

	return transitionsToState;
}

template<class SymbolType, class StateType >
ext::multimap < StateType, StateType > EpsilonNFA < SymbolType, StateType >::getEpsilonTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist" );

	ext::multimap < StateType, StateType > transitionsToState;

	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions )
		if ( transition.second == to && transition.first.second.is_epsilon ( ) )
			transitionsToState.insert ( transition.first.first, to );

	return transitionsToState;
}

template<class SymbolType, class StateType >
ext::multimap < ext::pair < StateType, SymbolType >, StateType > EpsilonNFA < SymbolType, StateType >::getSymbolTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + ext::to_string ( to ) + "\" doesn't exist" );

	ext::multimap < ext::pair < StateType, SymbolType >, StateType > transitionsToState;

	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions )
		if ( transition.second == to && ! transition.first.second.is_epsilon ( ) )
			transitionsToState.insert ( ext::pair < StateType, SymbolType > ( transition.first.first, transition.first.second.getSymbol ( ) ), to );

	return transitionsToState;
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::isEpsilonFree ( ) const {
	for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : transitions )
		if ( transition.first.second.is_epsilon ( ) )
			return false;

	return true;
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::isDeterministic ( ) const {
	if ( transitions.empty ( ) )
		return true;

	for ( auto iter = transitions.begin ( ); std::next ( iter ) != transitions.end ( ); ++ iter )
		if ( iter->first == std::next ( iter )->first )
			return false;

	return isEpsilonFree ( );
}

template<class SymbolType, class StateType >
bool EpsilonNFA < SymbolType, StateType >::isTotal ( ) const {
	return isDeterministic ( ) && transitions.size ( ) == getInputAlphabet ( ).size ( ) * getStates ( ).size ( );
}

template<class SymbolType, class StateType >
int EpsilonNFA < SymbolType, StateType >::compare ( const EpsilonNFA & other ) const {
	auto first = ext::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = ext::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	static ext::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template<class SymbolType, class StateType >
EpsilonNFA < SymbolType, StateType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace automaton */

namespace core {

/**
 * Helper class specifying constraints for the automaton's internal input alphabet component.
 *
 * \tparam SymbolType used for the terminal alphabet of the automaton.
 * \tparam StateType used for the terminal alphabet of the automaton.
 */
template<class SymbolType, class StateType >
class SetConstraint< automaton::EpsilonNFA < SymbolType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	/**
	 * Returns true if the symbol is still used in some transition of the automaton.
	 *
	 * \param automaton the tested automaton
	 * \param symbol the tested symbol
	 *
	 * \returns true if the symbol is used, false othervise
	 */
	static bool used ( const automaton::EpsilonNFA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : automaton.getTransitions ( ) )
			if ( ! transition.first.second.is_epsilon ( ) && ( transition.first.second.getSymbol ( ) == symbol ) )
				return true;

		return false;
	}

	/**
	 * Returns true as all symbols are possibly available to be elements of the input alphabet.
	 *
	 * \param automaton the tested automaton
	 * \param symbol the tested symbol
	 *
	 * \returns true
	 */
	static bool available ( const automaton::EpsilonNFA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	/**
	 * All symbols are valid as input symbols.
	 *
	 * \param automaton the tested automaton
	 * \param symbol the tested symbol
	 */
	static void valid ( const automaton::EpsilonNFA < SymbolType, StateType > &, const SymbolType & ) {
	}
};

/**
 * Helper class specifying constraints for the automaton's internal states component.
 *
 * \tparam SymbolType used for the terminal alphabet of the automaton.
 * \tparam StateType used for the terminal alphabet of the automaton.
 */
template<class SymbolType, class StateType >
class SetConstraint< automaton::EpsilonNFA < SymbolType, StateType >, StateType, automaton::States > {
public:
	/**
	 * Returns true if the state is still used in some transition of the automaton.
	 *
	 * \param automaton the tested automaton
	 * \param state the tested state
	 *
	 * \returns true if the state is used, false othervise
	 */
	static bool used ( const automaton::EpsilonNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > & transition : automaton.getTransitions ( ) )
			if ( transition.first.first == state || transition.second == state )
				return true;

		return false;
	}

	/**
	 * Returns true as all states are possibly available to be elements of the states.
	 *
	 * \param automaton the tested automaton
	 * \param state the tested state
	 *
	 * \returns true
	 */
	static bool available ( const automaton::EpsilonNFA < SymbolType, StateType > &, const StateType & ) {
		return true;
	}

	/**
	 * All states are valid as a state of the automaton.
	 *
	 * \param automaton the tested automaton
	 * \param state the tested state
	 */
	static void valid ( const automaton::EpsilonNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

/**
 * Helper class specifying constraints for the automaton's internal final states component.
 *
 * \tparam SymbolType used for the terminal alphabet of the automaton.
 * \tparam StateType used for the terminal alphabet of the automaton.
 */
template<class SymbolType, class StateType >
class SetConstraint< automaton::EpsilonNFA < SymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	/**
	 * Returns false. Final state is only a mark that the automaton itself does require further.
	 *
	 * \param automaton the tested automaton
	 * \param state the tested state
	 *
	 * \returns false
	 */
	static bool used ( const automaton::EpsilonNFA < SymbolType, StateType > &, const StateType & ) {
		return false;
	}

	/**
	 * Determines whether the state is available in the automaton's states set.
	 *
	 * \param automaton the tested automaton
	 * \param state the tested state
	 *
	 * \returns true if the state is already in the set of states of the automaton
	 */
	static bool available ( const automaton::EpsilonNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	/**
	 * All states are valid as a final state of the automaton.
	 *
	 * \param automaton the tested automaton
	 * \param state the tested state
	 */
	static void valid ( const automaton::EpsilonNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

/**
 * Helper class specifying constraints for the automaton's internal initial state element.
 *
 * \tparam SymbolType used for the terminal alphabet of the automaton.
 * \tparam StateType used for the terminal alphabet of the automaton.
 */
template<class SymbolType, class StateType >
class ElementConstraint< automaton::EpsilonNFA < SymbolType, StateType >, StateType, automaton::InitialState > {
public:
	/**
	 * Determines whether the state is available in the automaton's states set.
	 *
	 * \param automaton the tested automaton
	 * \param state the tested state
	 *
	 * \returns true if the state is already in the set of states of the automaton
	 */
	static bool available ( const automaton::EpsilonNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	/**
	 * All states are valid as an initial state of the automaton.
	 *
	 * \param automaton the tested automaton
	 * \param state the tested state
	 */
	static void valid ( const automaton::EpsilonNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

/**
 * Helper for normalisation of types specified by templates used as internal datatypes of symbols and states.
 *
 * \returns new instance of the automaton with default template parameters or unmodified instance if the template parameters were already the default ones
 */
template<class SymbolType, class StateType >
struct normalize < automaton::EpsilonNFA < SymbolType, StateType > > {
	static automaton::EpsilonNFA < > eval ( automaton::EpsilonNFA < SymbolType, StateType > && value ) {
		ext::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( value ).getInputAlphabet ( ) );
		ext::set < DefaultStateType > states = automaton::AutomatonNormalize::normalizeStates ( std::move ( value ).getStates ( ) );
		DefaultStateType initialState = automaton::AutomatonNormalize::normalizeState ( std::move ( value ).getInitialState ( ) );
		ext::set < DefaultStateType > finalStates = automaton::AutomatonNormalize::normalizeStates ( std::move ( value ).getFinalStates ( ) );

		automaton::EpsilonNFA < > res ( std::move ( states ), std::move ( alphabet ), std::move ( initialState ), std::move ( finalStates ) );

		for ( std::pair < ext::pair < StateType, common::symbol_or_epsilon < SymbolType > >, StateType > && transition : ext::make_mover ( std::move ( value ).getTransitions ( ) ) ) {
			DefaultStateType from = automaton::AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			common::symbol_or_epsilon < DefaultSymbolType > input = automaton::AutomatonNormalize::normalizeSymbolEpsilon ( std::move ( transition.first.second ) );
			DefaultStateType target = automaton::AutomatonNormalize::normalizeState ( std::move ( transition.second ) );

			res.addTransition ( std::move ( from ), std::move ( input ), std::move ( target ) );
		}

		return res;
	}
};

} /* namespace core */

extern template class automaton::EpsilonNFA < >;

#endif /* EPSILON_NFA_H_ */
